"""
This module provides test data for the telemetry storage tests.
"""

from test_data import DEFAULT_TELEMETRY_MODEL_PARAMS  # noqa: F401
from test_data import DEFAULT_TELEMETRY_PARAMS  # noqa: F401
